
/* The main routine for AP238 export (see ap238export() see below)

   1. Set everything up
   2. Save all the tools to the STEP-NC file (see save_all_milling_tools() below)
   3. Save all the operations to the STEP-NC file (see save_all_operations() below)

   Mastercam calls the main routine (ap238export()) using the hooks and gizmos of a Windows DLL in the file stdafx.cpp.
   The STEP-NC data is made by calling the functions in the OperationMaker, FeatureMaker and Toolmaker libraries.

   Documentation for these libraries can be found on the the STEP Tools web site at.

   http://www.steptools.com/support/stdev_docs/operation_maker/
   http://www.steptools.com/support/stdev_docs/feature_maker/
   http://www.steptools.com/support/stdev_docs/tool_maker/

   */

#include "stdafx.h"

#include "m_core.h"
#include "m_mastercam.h"
#include "m_mill.h"

#include "ap238export.h"

// pfLog is used for optional STEP-NC trace functionality in all the maker libraries
FILE* pfLog = NULL;

// mcLog is used for optional Mastercam trace functionality in ap238export
FILE* mcLog = NULL;

// Calls to this function can be eliminated from code by unsetting #MY_TRACE in ToolMaker.h
void ap238_function_trace(const char * fmt, ...)
{
#ifdef MC_TRACE
    if (mcLog == NULL) 
	return;
    va_list ap;
    va_start(ap, fmt);

    char buff[10240];
    char * c = buff;
    vsprintf (c, fmt, ap);

    va_end(ap);

    static int count = 0;
//    printf ("Operation_maker Call#%d: %s\n", count, message);
    fprintf (mcLog, "MC EXPORT\tCall#%d: %s\n", count, buff);
    count++;   
#endif
}

// Error reports can go to a message box, the trace file or nowhere depending on setting of #MY_REPORT and #MY_TRACE
void ap238_report_error(const char * fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);

    char buff[10240];
    char * c = buff;
    vsprintf (c, fmt, ap);

    va_end(ap);

    // Windows gadget to show a warning box requires windows.h
#ifdef MC_REPORT
    MessageBox (NULL, buff, "WARNING", MB_ICONSTOP);
#endif
#ifdef MC_TRACE
    if (mcLog == NULL) 
	return;
    fprintf (mcLog, "MC EXPORT\tERROR: %s\n", buff);
#endif
}


// Called once to setup functionality
void initialize_ap238export()
{
    MC_TRACE ("initialize_ap238export");

    static RoseBoolean initialized = ROSE_FALSE;

    if (initialized)
	return;

    ROSE.quiet (ROSE_TRUE);		// turn off startup messages

    stplib_init();

    ST_MODULE_FORCE_LOAD();    
    initialized = ROSE_TRUE;
}


// The main routine for the STEP-NC export
extern void ap238export() {

    CString fnme = "export.stpnc";
    CString ap238_trace_file;
    CString mc_trace_file;

    Milling_machine_cutting_tool_IF *step_tools[255]; 
    tp_tool mastercam_tools[255];
    unsigned tool_count = -1;


    CFileDialog dlgSave(FALSE, ".stpnc", fnme, OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT,
			"STEP Files (*.stpnc;*.stp;*.step;*.p21;*.238)|*.stpnc;*.stp;*.step;*.p21;*.238|All Files (*.*)|*.*||", NULL);
    try {
	if (IDOK != dlgSave.DoModal()) {
		    return;
	}

	if (NEED_MC_TRACE_FILE) {
	    mc_trace_file = dlgSave.GetFolderPath ();
	    mc_trace_file += "/mc_trace.txt";
	    mcLog = fopen(mc_trace_file, "w+t");
	}
	else
	    mcLog = NULL;

	if (NEED_AP238_TRACE_FILE) {
	    ap238_trace_file = dlgSave.GetFolderPath ();
	    ap238_trace_file += "/ap238_trace.txt";
	    pfLog = fopen(ap238_trace_file, "w+t");
	}
	else
	    pfLog = NULL;

	initialize_ap238export();
	// MessageBox (0, "First Hello from Mastercam!", "", MB_OK);		// uncomment this if you want to be sure the system is starting
	RoseDesign *des = ROSE.newDesign (dlgSave.GetPathName());

	Project * project = new_project (des, dlgSave.GetFileName ());
	Workplan *main_plan = main_workplan (project, dlgSave.GetFileName ());
	save_all_milling_tools (des, tool_count, step_tools, mastercam_tools);
	save_all_milling_operations (main_plan, tool_count, step_tools, mastercam_tools);
	ARMsave (des);

	if (pfLog) {
	    fclose (pfLog);
	    pfLog = NULL;
	}

	if (mcLog) {
	    fclose (mcLog);
	    mcLog = NULL;
	}

	MessageBox (0, "Mastercam STEP-NC export done!", "", MB_OK);

    }

    catch (...) {
	if (pfLog) {
	    fclose (pfLog);
	    pfLog = NULL;
	}

	if (mcLog) {
	    fclose (mcLog);
	    mcLog = NULL;
	}
	AfxMessageBox("Can't export toolpaths.", MB_OK | MB_ICONSTOP);
    }

    return;
}

// write all the tools to the STEP-NC file using the ToolMaker functions see http://www.steptools.com/support/stdev_docs/tool_maker/
int save_all_milling_tools (RoseDesign *des, unsigned &tool_count, Milling_machine_cutting_tool_IF *step_tools[], tp_tool mastercam_tools[])
{
    tp_tool tl;
    MC_BOOL bResult = true;
    boolean foundOne = false;					// can be used to decide if need to search lathe tools 
    DB_LIST_ENT_PTR ptrTool;
    unsigned i;

    Milling_machine_cutting_tool_IF *tool;

    StixUnit length_unit, angle_unit;
    if (IsEnglish())
    {
	MC_TRACE("Writing tool requirements in Inches and Degrees");
	length_unit = stixunit_in;
	angle_unit = stixunit_deg;
    }
    else
    {
	MC_TRACE("Writing tool requirements in Millimeters and Degrees");
	length_unit = stixunit_in;
	angle_unit = stixunit_deg;
    }

    tool_count = 0;

    // Enumerate tools
    for (i = 1; i < 256; i++) {
	memset(&tl, 0, sizeof(tp_tool));

	tl.op.slot = i;
	tool_manager(&tl, TLMGR_GET, &ptrTool, &bResult);

	if (!bResult) {
	    continue;
	}
	else {
	    foundOne = TRUE;
	}

	int hand_of_cut = tl.spindle_rot;

	char buf[12];
	sprintf (buf, "%d", tl.op.tlno);		// Tool number is used as tool id

	if (tl.op.type == 10) { // end mill flat
	    Endmill *end = new_endmill_tool (des, buf, tl.op.dia, tl.holder_length + tl.oa_length, tl.oa_length, tl.flute_length, tl.op.n_flutes, 0, length_unit, angle_unit);
	    MC_TRACE ("");
	    MC_TRACE ("Saved Endmill Tool %d MCAM id = %d with identifier %s and type = %d", tl.op.slot, tl.op.tlno, tl.op.comment, tl.op.type);
	    MC_TRACE ("Diameter = %f, Overall length = %f, functional length = %f, flute length = %f", tl.op.dia, tl.holder_length + tl.oa_length, tl.oa_length, tl.flute_length);
	    MC_TRACE ( "Num flutes = %d Taper angle (saved as 0) = %f", tl.op.n_flutes, tl.op.tip_angle);
	    if (tl.op.rad_type) {
		set_tool_radius (end, tl.op.crad, length_unit);
		MC_TRACE("Corner radius = %f, rad_type = %d", tl.op.crad, tl.op.rad_type);
	    }
	    tool = end;
	}
	else if (tl.op.type == 12) { // chamfer tool (tool tip angle not 0)
	    Endmill *end = new_endmill_tool (des, buf, tl.op.dia, tl.holder_length + tl.oa_length, tl.oa_length, tl.flute_length, tl.op.n_flutes, tl.op.tip_angle, length_unit, angle_unit);
	    MC_TRACE ("");
	    MC_TRACE("Saved Chamfer Endmill Tool %d MCAM id = %d with identifier %s and type = %d", tl.op.slot, tl.op.tlno, tl.op.comment, tl.op.type);
	    MC_TRACE("Diameter = %f, Overall length = %f, functional length = %f, flute length = %f", tl.op.dia, tl.holder_length + tl.oa_length, tl.oa_length, tl.flute_length);
	    MC_TRACE("Num flutes = %d Taper angle = %f", tl.op.n_flutes, tl.op.tip_angle);
	}
	else if (tl.op.type == 13) {
	    Facemill *face = new_facemill_tool (des, buf, tl.op.dia, tl.holder_length + tl.oa_length, tl.oa_length, tl.flute_length, tl.op.n_flutes, tl.op.tip_angle, length_unit, angle_unit);
	    MC_TRACE ("");
	    MC_TRACE("Saved Facemill Tool %d MCAM id = %d with identifier %s and type = %d", tl.op.slot, tl.op.tlno, tl.op.comment, tl.op.type);
	    MC_TRACE("Diameter = %f, Overall length = %f, functional length = %f, flute length = %f", tl.op.dia, tl.holder_length + tl.oa_length, tl.oa_length, tl.flute_length);
	    MC_TRACE("Num flutes = %d Taper angle = %f", tl.op.n_flutes, tl.op.tip_angle);
	    if (tl.op.rad_type) {
		set_tool_radius (face, tl.op.crad, length_unit);
		MC_TRACE("Corner radius = %f, rad_type = %d", tl.op.crad, tl.op.rad_type);
	    }
	    tool = face;
	}
	else if (tl.op.type == 2) {
	    Spotdrill *spot = new_spot_drill (des, buf, tl.op.dia, tl.holder_length + tl.oa_length, tl.oa_length, tl.flute_length, tl.op.tip_angle, length_unit, angle_unit);
	    MC_TRACE ("");
	    MC_TRACE("Saved Spot Drill Tool %d MCAM id = %d with identifier %s and type = %d", tl.op.slot, tl.op.tlno, tl.op.comment, tl.op.type);
	    MC_TRACE("Diameter = %f, Overall length = %f, functional length = %f, flute length = %f", tl.op.dia, tl.holder_length + tl.oa_length, tl.oa_length, tl.flute_length);
	    MC_TRACE("tip angle = %f", tl.op.tip_angle);
	    tool = spot;
	}
	else if (tl.op.type == 3) {
	    Twist_drill *twist = new_twist_drill (des, buf, tl.op.dia, tl.holder_length + tl.oa_length, tl.oa_length, tl.flute_length, tl.op.tip_angle, length_unit, angle_unit);
	    MC_TRACE ("");
	    MC_TRACE("Saved (Twist) Drill Tool %d MCAM id = %d with identifier %s and type = %d", tl.op.slot, tl.op.tlno, tl.op.comment, tl.op.type);
	    MC_TRACE("Diameter = %f, Overall length = %f, functional length = %f, flute length = %f", tl.op.dia, tl.holder_length + tl.oa_length, tl.oa_length, tl.flute_length);
	    MC_TRACE("tip angle = %f", tl.op.tip_angle);
	    tool = twist;
	}
	else if (tl.op.type == 4) {
	    Tapping_cutting_tool *tap = new_tapping_tool (des, buf, tl.op.dia, tl.holder_length + tl.oa_length, tl.oa_length, tl.flute_length, "metric", tl.op.thds, tl.op.dia -tl.op.thds, length_unit);
	    set_tool_hand_of_cut (tap, EN_RIGHT);
	    MC_TRACE ("");
	    MC_TRACE("Saved Tapping RH Tool %d MCAM id = %d with identifier %s and type = %d", tl.op.slot, tl.op.tlno, tl.op.comment, tl.op.type);
	    MC_TRACE("Diameter = %f, Overall length = %f, functional length = %f, flute length = %f", tl.op.dia, tl.holder_length + tl.oa_length, tl.oa_length, tl.flute_length);
	    MC_TRACE("thread_pitch = %f", tl.op.thds);
	    tool = tap;
	}
	else if (tl.op.type == 5) {
	    Tapping_cutting_tool *tap = new_tapping_tool (des, buf, tl.op.dia, tl.holder_length + tl.oa_length, tl.oa_length, tl.flute_length, "metric", tl.op.thds, tl.op.dia -tl.op.thds, length_unit);
	    set_tool_hand_of_cut (tap, EN_LEFT);
	    MC_TRACE ("");
	    MC_TRACE("Saved Tapping LH Tool %d MCAM id = %d with identifier %s and type = %d", tl.op.slot, tl.op.tlno, tl.op.comment, tl.op.type);
	    MC_TRACE("Diameter = %f, Overall length = %f, functional length = %f, flute length = %f", tl.op.dia, tl.holder_length + tl.oa_length, tl.oa_length, tl.flute_length);
	    MC_TRACE("thread_pitch = %f", tl.op.thds);
	    tool = tap;
	}
	else if (tl.op.type == 7) {
	    Rotating_boring_cutting_tool *bore = new_boring_tool (des, buf, tl.op.dia, tl.holder_length + tl.oa_length, tl.oa_length, tl.flute_length, length_unit);
	    MC_TRACE ("");
	    MC_TRACE("Saved Bore Tool %d MCAM id = %d with identifier %s and type = %d", tl.op.slot, tl.op.tlno, tl.op.comment, tl.op.type);
	    MC_TRACE("Diameter = %f, Overall length = %f, functional length = %f, flute length = %f", tl.op.dia, tl.holder_length + tl.oa_length, tl.oa_length, tl.flute_length);
	    tool = bore;
	}
	else {
	    // as per note from magnus overall assembly length in MCam is holder_length + oa_length
	    User_defined_milling_tool *user = new_apt_tool (des, buf, tl.op.dia, tl.op.crad, 0, 0, 0, 0, tl.holder_length + tl.oa_length, length_unit, angle_unit);
	    MC_TRACE ("");
	    MC_TRACE("Saved as APT Tool %d MCAM id = %d with identifier %s and type = %d and tip = %f", tl.op.slot, tl.op.tlno, tl.op.comment, tl.op.type, tl.op.tip_angle);
	    tool = user;
	}

	// add additional data about tool material and feeds and speeds if available
	if (tl.op.matl == 1) {
	    set_tool_cutting_material (tool, "HSS", "Mastercam");
	    MC_TRACE("Material code = %d  value = HSS ", tl.op.matl);
	}
	else if (tl.op.matl == 2) {
	    set_tool_cutting_material (tool, "Carbide", "Mastercam");
	    MY_TRACE(pfLog, "Material code = %d  value = Carbide ", tl.op.matl);
	}
	else if (tl.op.matl == 3) {
	    set_tool_cutting_material (tool, "TiCoated", "Mastercam");
	    MC_TRACE("Material code = %d  value = TiCoated ", tl.op.matl);
	}
	else if (tl.op.matl == 3) {
	    set_tool_cutting_material (tool, "Ceramic", "Mastercam");
	    MC_TRACE("Material code = %d  value = Ceramic ", tl.op.matl);
	}
	else
	    MC_TRACE("Material code = %d  value = Unknown ", tl.op.matl);

	if (tl.op.feed != 0) {
	    if (length_unit == stixunit_in) {
		set_tool_recommended_feed (tool, tl.op.feed, stixunit_ips);
	    }
	    if (length_unit == stixunit_mm) {
		set_tool_recommended_feed (tool, tl.op.feed, stixunit_ipm);
	    }
	    MC_TRACE("Recommended Feed = %f", tl.op.feed);
	}
	else
	    MC_TRACE("No feed recommendation set (Feed = %f)", tl.op.feed);

	if (tl.op.rpm != 0) {
	    set_tool_recommended_speed (tool, tl.op.rpm, stixunit_rpm);
	    MC_TRACE("Recommended Speed = %d", tl.op.rpm);
	}
	else
	    MC_TRACE("No speed recommendation set (Speed = %d)", tl.op.rpm);

	step_tools[tool_count] = tool;
	mastercam_tools[tool_count] = tl;
	tool_count++;
	if (tool_count >= 255) {
	    MC_TRACE("Too many tools in Mastercam project (limit is 255 - contact STEP Tools)");
	    return 0;
	}

   }

    return foundOne;
}

#include <cmath>

BOOL VEq(p_3d v1, p_3d v2)
{
    return (v1[0] == v2[0] &&
	    v1[1] == v2[1] &&
	    v1[2] == v2[2]);
}

BOOL MEq (x_matrix m1, x_matrix m2) {
    return VEq (m1.t, m2.t) && VEq (m1.r[0], m2.r[0]) && VEq (m1.r[1], m2.r[1]) && VEq (m1.r[2], m2.r[2]);
}

// Write all the operations to the STEP-NC file using the OperationMaker functions see http://www.steptools.com/support/stdev_docs/operation_maker/
int save_all_milling_operations (Workplan *plan, unsigned tool_count, Milling_machine_cutting_tool_IF *step_tools[], tp_tool mastercam_tools[])
{
    INT_PTR OpListSize = TpMainOpMgr.GetMainOpList().GetSize();

    if (OpListSize == 0) {
	MC_TRACE("Empty Operation list\n");
	return 0;
    }

    operation *opl = NULL;              
    op_information info;
    opl = TpMainOpMgr.GetMainOpList()[0];
    op_info (opl->op_idn, TpMainOpMgr.GetMainOpList(), &info);

    x_matrix last_x_matrix;
    memset (last_x_matrix.t, 0, sizeof (p_3d));
    memset (last_x_matrix.r, 0, sizeof (matrix_33));

    Workplan *ws_plan = plan;
    int setup_number = 1;

    if (info.is_lathe) {
	MC_TRACE("This Mastercam file describes operations for a LATHE\n");
	return 0;
    }


    CString message;
    CString caption = "STEP-NC Export";
    message.Format ("Export all data with no more questions");
    long main_q = AfxMessageBox(message, MB_YESNO | MB_ICONINFORMATION);
    if (main_q == IDYES) {
	    MY_TRACE(pfLog, "Exporting all data with no questions asked\n");
    }

    MC_BOOL bResult;
    nci_bin n;
    long fpos;

    // Rewind to the beginning of the NCI file
    nci_manager(opl->op_idn, NCIMGR_FSTART, &n, &fpos, &bResult);

    CStringArray old_names, new_names;
    for (INT_PTR OpCount = 0; OpCount < OpListSize; ++OpCount)
    {      
	MC_TRACE ("");
	MC_TRACE ("Top of main MASTERCAM LOOP Count = %d of %d", OpCount, OpListSize);
	opl = TpMainOpMgr.GetMainOpList()[OpCount];
	char opDescription[80] = "";
	op_make_description (opl, FALSE, opDescription, 80);
	MC_TRACE ("opDescription = %s", opDescription);

	// workplan hierarchy for the next operation
	GetOpParents(opl, &new_names); // Get the "parents" of this operation
	if (OpCount == 0) {				    // First operation create any necessary workplans below the main workplan
	    INT_PTR NamesSize = new_names.GetSize ();
	    MC_TRACE("Mastercam operation stack has height %d", NamesSize);
	    old_names.Add (new_names[0]);
	    // Edited on July 2 2012 to eliminate "Toolpath Group" nested workplan - was NameSize - 2 now NameSize - 3
	    // First name is machine group, next is toolpath group
	    MC_TRACE ("Main workplan name set to %s", new_names[1]);
	    plan->put_its_id (new_names[1]);
	    for (INT_PTR i = 2; i < NamesSize; i++) {
		MC_TRACE("Making nested workplan called %s", new_names[i]);
			plan = new_workplan (plan, new_names[i]);
//			g_Apt.NestWorkplan (new_names[i]);
	    }
	}
	else {						    // Second and subsequent operations: has the workplan hierarchy grown or shrunk?
	    INT_PTR NamesSize = new_names.GetSize ();
	    INT_PTR OldNamesSize = old_names.GetSize ();
	    MC_TRACE("Mastercam operation stack: new height %d old height %d", NamesSize, OldNamesSize);
	    if (NamesSize > OldNamesSize) {
		for (INT_PTR i = OldNamesSize; i < NamesSize; i++) {
		    MC_TRACE("Starting nested workplan called %s", new_names[i]);
//		    new_workplan (parent, new_names[i]);
		}
	    }
	    else if (NamesSize < OldNamesSize) {
		for (INT_PTR i = OldNamesSize - 1; i >= NamesSize; i--) {
		    MC_TRACE("Ending nested workplan called %s", old_names[i]);
//		    end_workplan ();
		}
	    }
	}

	// remember names for next iteration of main loop
	old_names.RemoveAll();
	INT_PTR NamesSize = new_names.GetSize ();
	for (INT_PTR i = 0; i < NamesSize; i++) {
		old_names.Add (new_names[i]);
	}


	if (main_q == IDNO) {
		CString message;
		CString caption = "STEP-NC Export";
		message.Format ("Export Operation '%s' (%d of %d)", opDescription, OpCount + 1, OpListSize);
		long res = AfxMessageBox(message, MB_YESNO | MB_ICONINFORMATION);
		if (res == IDNO) {
			MY_TRACE(pfLog, "Skipping operation %d of %d at user request\n", OpCount + 1, OpListSize);
			char buf[255];		   
			sprintf (buf, "Operation '%s' skipped at user request", opDescription);
			new_workingstep (plan, buf);
			continue;
		}
	}

	x_matrix new_x_matrix;
	memcpy (new_x_matrix.t, opl->WCS.origin, sizeof (p_3d));
	memcpy (new_x_matrix.r, opl->WCS.m, sizeof (matrix_33));

	if (opl->WCS.on)
	    MC_TRACE("WCS is on for operation called %s with comment %s", opDescription, opl->comment);

//	Neither of the below methods for determining whether a new setup is satisfactory - both are too aggressive and create too many setups
//	if (opl->WCS.on && !MEq (new_x_matrix, last_x_matrix)) { // New setup on any WCS change
	if (opl->WCS.on && !VEq (new_x_matrix.t, last_x_matrix.t)) { // New setup on placement change
	    char setup_name[100];
	    sprintf (setup_name, "Setup %d", setup_number++);
	    MC_TRACE("Making new workplan for new Setup called %s with comment %s", setup_name, opl->comment);
	    ws_plan = new_workplan (plan, setup_name);

	    StixMtrx M;
	    M.w(StixVctr (new_x_matrix.t[0], new_x_matrix.t[1], new_x_matrix.t[2]));
	    M.x(StixVctr (new_x_matrix.r[0][0], new_x_matrix.r[0][1], new_x_matrix.r[0][2]));
	    M.y(StixVctr (new_x_matrix.r[1][0], new_x_matrix.r[1][1], new_x_matrix.r[1][2]));
	    M.z(StixVctr (new_x_matrix.r[2][0], new_x_matrix.r[2][1], new_x_matrix.r[2][2]));

	    MC_TRACE("Setup Origin (%f, %f, %f)", new_x_matrix.t[0], new_x_matrix.t[1], new_x_matrix.t[2]);
	    MC_TRACE("Setup X (%f, %f, %f)", new_x_matrix.r[0][0], new_x_matrix.r[0][1], new_x_matrix.r[0][2]);
	    MC_TRACE("Setup Y (%f, %f, %f)", new_x_matrix.r[1][0], new_x_matrix.r[1][1], new_x_matrix.r[1][2]);
	    MC_TRACE("Setup Z (%f, %f, %f)", new_x_matrix.r[2][0], new_x_matrix.r[2][1], new_x_matrix.r[2][2]);

	    workplan_setup (ws_plan, M);
	    memcpy (last_x_matrix.t, new_x_matrix.t, sizeof (p_3d));
	    memcpy (last_x_matrix.r, new_x_matrix.r, sizeof (matrix_33));
	}

	MC_TRACE("Making new workingstep called %s with comment %s", opDescription, opl->comment);
	Machining_workingstep *ws;
	if (strlen (opDescription) > 0)
	    ws = new_workingstep (ws_plan, opDescription);
	else if (opl->opcode == TP_DRILL)
	    ws = new_workingstep (ws_plan, "drill"); 
	else
	    ws = new_workingstep (ws_plan, "mill");

	save_ws_toolpaths (ws, opl, tool_count, step_tools, mastercam_tools);
   }

    return 0;
}

// save one operation to the export file
// to keep it simple this translator works just for the operations required by Boxy
int save_ws_toolpaths (Machining_workingstep *ws, operation *op, unsigned tool_count, Milling_machine_cutting_tool_IF *step_tools[], tp_tool mastercam_tools[]) 
{
    MC_BOOL bResult;
    MC_BOOL bFirstMove = TRUE;
    MC_BOOL bInRapidMode = FALSE;

    p_3d ptStart;					// start point for the next operation
    memset(&ptStart, 0, 3 * sizeof(double));

    double hole_z_top, hole_z_bot;			// hole top and bottom for hole patterns

    nci_bin n;
    long fpos;

    double feed, oldFeed;				// oldFeed, oldRPM used to signal requirement for new toolpath
    int RPM, oldRPM;

    Milling_technology *tech = NULL;			// STEP-NC technology (feed and speed)
    Milling_machine_functions *funs = NULL;		// STEP-NC functions (coolant, mist)

    StixUnit length_unit, angle_unit, speed_unit, feed_unit;
    if (IsEnglish())
    {
	MC_TRACE("Writing operation in Inches and Degrees");
	length_unit = stixunit_in;
	angle_unit = stixunit_deg;
	speed_unit = stixunit_rpm;
	feed_unit = stixunit_ips;
    }
    else
    {
	MC_TRACE("Writing operation in Millimeters and Degrees");
	length_unit = stixunit_mm;
	angle_unit = stixunit_deg;
	speed_unit = stixunit_rpm;
	feed_unit = stixunit_mmpm;
    }


    // Move to the beginning of the operation stream
    nci_manager(op->op_idn, NCIMGR_RD_SOS, &n, &fpos, &bResult);

    if (!bResult) {
	MC_TRACE("Mastercam nci_manager failure - aborting workingstep");
	return 0;
    }

    CString cmnt = op->comment;

    if (mastercam_tools[op->tl.tlno - 1].spindle_rot)
	oldRPM= RPM = op->tl.rpm;
    else
	oldRPM = RPM = -op->tl.rpm;

    oldFeed = feed = mastercam_tools[op->tl.tlno - 1].op.feed;

    const char *coolant_on, *mist_on, *through_spindle_on, *chip_removal_on;
    if (mastercam_tools[op->tl.tlno - 1].op.coolant)
	coolant_on = EN_COOLANT_ON;
    else
	coolant_on = EN_COOLANT_OFF;

    if (mastercam_tools[op->tl.tlno - 1].op.coolant == 2)
	mist_on = EN_MIST_ON;
    else
	mist_on = EN_MIST_OFF;

    if (mastercam_tools[op->tl.tlno - 1].op.coolant == 3)
	through_spindle_on = EN_THROUGH_SPINDLE_COOLANT_ON;
    else
	through_spindle_on = EN_THROUGH_SPINDLE_COOLANT_OFF;

    chip_removal_on = EN_CHIP_REMOVAL_ON;			// where in Mastercam?

    tech = find_or_make_technology (ws->getRootObject()->design(), feed, feed_unit, RPM, speed_unit);
    funs = find_or_make_functions (ws->getRootObject()->design(), coolant_on, mist_on, through_spindle_on, chip_removal_on);

    MC_TRACE ("New freefrom operation tool no = %d, tool_type = %s", op->tl.tlno, step_tools[op->tl.tlno - 1]->getModuleName());
    Machining_operation_IF *step_op = new_freeform_operation (ws, step_tools[op->tl.tlno - 1], tech, funs);
    set_toolpath_units (step_op, length_unit, angle_unit);
    Toolpath_IF *path = NULL;

    // -----------------------------------------------------------
    // BEGIN reading operations in a loop from Mastercam NCI data
    // -----------------------------------------------------------
    while (bResult) {

	// Read from the Mastercam operation stream
	nci_manager(op->op_idn, NCIMGR_RD, &n, &fpos, &bResult);
	MC_TRACE("");
	MC_TRACE("HERE IS THE NEW GCODE: %d", n.gcode);

	if (!bResult || n.op_idn != op->op_idn) {
	    if (!bResult)
		MC_TRACE ("Mastercam nci manager finished operation - breaking loop");
	    else
		MC_TRACE("Mastercam operation number changed from %d to %d - breaking loop", op->op_idn, n.op_idn);
	    return 0;
	}

	if (n.gcode == 20100) {
	    long ltcode = n.u.l20100.lToolNum;
	    MC_TRACE("LATHE TOOL IN SLOT: %d Number %d", n.u.l20100.uSlot, ltcode);
	    MC_TRACE("Breaking loop");
	    return 0;
	}

	// MC Toolplane used to set STEP-NC workingstep toolpath orientation
	if (n.gcode == 1027) { // gcode 1027 = 1013 (tplane origin) + 1014 (tplane orientation)
	    MC_TRACE("Origin found (%f, %f, %f)", n.u.m1027.wcs_origin[0], n.u.m1027.wcs_origin[1], n.u.m1027.wcs_origin[2]);
	    MC_TRACE("X found (%f, %f, %f)", n.u.m1027.wcs_m[0][0], n.u.m1027.wcs_m[0][1], n.u.m1027.wcs_m[0][2]);
	    MC_TRACE("Y found (%f, %f, %f)", n.u.m1027.wcs_m[1][0], n.u.m1027.wcs_m[1][1], n.u.m1027.wcs_m[1][2]);
	    MC_TRACE("Z found (%f, %f, %f)", n.u.m1027.wcs_m[2][0], n.u.m1027.wcs_m[2][1], n.u.m1027.wcs_m[2][2]);

	    p_3d pt_origin;
	    x_matrix mXform2;
	    memset(&mXform2, 0, sizeof(x_matrix));
	    memcpy(mXform2.r, n.u.m1027.wcs_m, 9 * sizeof(double));

	    memcpy (pt_origin, n.u.m1027.wcs_origin, 3 * sizeof(double));
	    xform_pt(pt_origin, &mXform2);
	    MC_TRACE("Origin transformed for STEP-NC (%f, %f, %f)", pt_origin[0], pt_origin[1], pt_origin[2]);

	    StixMtrx M;
	    M.w(StixVctr (pt_origin[0], pt_origin[1], pt_origin[2]));
	    M.x(StixVctr (n.u.m1027.wcs_m[0][0], n.u.m1027.wcs_m[0][1], n.u.m1027.wcs_m[0][2]));
	    M.y(StixVctr (n.u.m1027.wcs_m[1][0], n.u.m1027.wcs_m[1][1], n.u.m1027.wcs_m[1][2]));
	    M.z(StixVctr (n.u.m1027.wcs_m[2][0], n.u.m1027.wcs_m[2][1], n.u.m1027.wcs_m[2][2]));
	    set_workingstep_toolpath_orientation (ws, M);

	}

// --------------------------------------------------------
// BEGIN toolpath extraction 
// --------------------------------------------------------

	// Now act based on the operaton type (G-code)
	if (n.gcode == 0) {// rapid 

	    if (n.u.m0.ccomp > 0)	    // use this code to control surface normal curve
		MC_TRACE("WARNING: Cutter compensation ignored");

	    if (op->opcode == TP_POINT) {   // Assume it's probing (NIST used this type of op)
		MC_TRACE("WARNING: Probing TP Point ignored");
	    }

	    // special case for first move in an operation
	    if (bFirstMove) {
		path = new_toolpath_rapid (step_op, EN_TRAJECTORY_PATH, cmnt);
		MC_TRACE("Starting op with rapid path at X%f Y%f Z%f", n.u.m0.ep1[0], n.u.m0.ep1[1], n.u.m0.ep1[2]);
		start_tool_tip_curve (path,  n.u.m0.ep1[0], n.u.m0.ep1[1], n.u.m0.ep1[2]);
		set_toolpath_functions (path, funs);
		set_toolpath_technology (path, tech);
		oldFeed = feed;
		oldRPM = RPM;
		bFirstMove = FALSE;
		bInRapidMode = TRUE;
		memcpy(ptStart, n.u.m0.ep1, 3 * sizeof(double));
		continue;
	    }

	    // change in feed or technology then start new path
	    if (oldFeed != feed || oldRPM != RPM || !bInRapidMode) {
		path = new_toolpath_rapid (step_op, EN_TRAJECTORY_PATH, cmnt);
		MC_TRACE("Starting new rapid path at X%f Y%f Z%f", ptStart[0], ptStart[1], ptStart[2]);
		start_tool_tip_curve (path,  ptStart[0], ptStart[1], ptStart[2]);
		set_toolpath_functions (path, funs);
		set_toolpath_technology (path, tech);
		oldFeed = feed;
		oldRPM = RPM;
		bInRapidMode = TRUE;
	    }

	    // Rapid move to next point
	    goto_xyz (path, n.u.m0.ep1[0], n.u.m0.ep1[1], n.u.m0.ep1[2]);
	    MC_TRACE("S%d, F%f", op->tl.rpm, n.u.m0.feed);
	    MC_TRACE("G0 X%f Y%f Z%f", n.u.m0.ep1[0], n.u.m0.ep1[1], n.u.m0.ep1[2]);
	    memcpy(ptStart, n.u.m0.ep1, 3 * sizeof(double));
	}

	else if (n.gcode == 1) {    // feed

	    if (op->tl.use_css) {
		    MC_TRACE("WARNING: Ignored CSS feed %f max %d\n", fabs(n.u.m1.feed), op->tl.max_ss);
	    }

	    feed = n.u.m1.feed;
	    tech = find_or_make_technology (ws->getRootObject()->design(), feed, feed_unit, RPM, speed_unit);
	    funs = find_or_make_functions (ws->getRootObject()->design(), coolant_on, mist_on, through_spindle_on, chip_removal_on);

	    if (n.u.m1.ccomp > 0)	    // use this code to control surface normal curve
		MC_TRACE("WARNING: Cutter compensation ignored");

	    // Special case for first move in the operation
	    if (bFirstMove) {
		path = new_toolpath (step_op, EN_TRAJECTORY_PATH, cmnt);
		MC_TRACE("Starting op with regular path at X%f Y%f Z%f", n.u.m1.ep1[0], n.u.m1.ep1[1], n.u.m1.ep1[2]);
		start_tool_tip_curve (path,  n.u.m1.ep1[0], n.u.m1.ep1[1], n.u.m1.ep1[2]);
		set_toolpath_functions (path, funs);
		set_toolpath_technology (path, tech);
		oldFeed = feed;
		oldRPM = RPM;
		bFirstMove = FALSE;
		memcpy(ptStart, n.u.m1.ep1, 3 * sizeof(double));
		continue;
	    }

	    // change in feed or technology then start new path
	    if (oldFeed != feed || oldRPM != RPM || bInRapidMode) {
		path = new_toolpath (step_op, EN_TRAJECTORY_PATH, cmnt);
		MC_TRACE("Starting path for new feed %f and speed %d at X%f Y%f Z%f", feed, RPM, ptStart[0], ptStart[1], ptStart[2]);
		start_tool_tip_curve (path,  ptStart[0], ptStart[1], ptStart[2]);
		set_toolpath_functions (path, funs);
		set_toolpath_technology (path, tech);
		oldFeed = feed;
		oldRPM = RPM;
		bInRapidMode = FALSE;
	    }

	    // move to next point
	    goto_xyz (path, n.u.m1.ep1[0], n.u.m1.ep1[1], n.u.m1.ep1[2]);
    	    MC_TRACE("S%d, F%f", op->tl.rpm, n.u.m1.feed);
	    MC_TRACE("G1 X%f Y%f Z%f", n.u.m1.ep1[0], n.u.m1.ep1[1], n.u.m1.ep1[2]);
	    memcpy(ptStart, n.u.m1.ep1, 3 * sizeof(double));
	}
					
	else if (n.gcode == 2 || n.gcode == 3) { // CW/CCW arc
	    bInRapidMode = FALSE;

	    feed = n.u.m2.feed;
	    tech = find_or_make_technology (ws->getRootObject()->design(), feed, feed_unit, RPM, speed_unit);
	    funs = find_or_make_functions (ws->getRootObject()->design(), coolant_on, mist_on, through_spindle_on, chip_removal_on);

	    if (bFirstMove) {
		path = new_toolpath (step_op, EN_TRAJECTORY_PATH, cmnt);
		MC_TRACE("Starting op with curve path at X%f Y%f Z%f", n.u.m1.ep1[0], n.u.m1.ep1[1], n.u.m1.ep1[2]);
		start_tool_tip_curve (path,  n.u.m1.ep1[0], n.u.m1.ep1[1], n.u.m1.ep1[2]);
		set_toolpath_functions (path, funs);
		set_toolpath_technology (path, tech);
		oldFeed = feed;
		oldRPM = RPM;
		bFirstMove = FALSE;
		memcpy(ptStart, n.u.m1.ep1, 3 * sizeof(double));
		continue;
	    }

	    // change in feed or technology then start new path
	    if (oldFeed != feed || oldRPM != RPM || bInRapidMode) {
		path = new_toolpath (step_op, EN_TRAJECTORY_PATH, cmnt);
		MC_TRACE("Starting curve path for new feed %f and speed %d at X%f Y%f Z%f", feed, RPM, ptStart[0], ptStart[1], ptStart[2]);
		start_tool_tip_curve (path,  ptStart[0], ptStart[1], ptStart[2]);
		set_toolpath_functions (path, funs);
		set_toolpath_technology (path, tech);
		oldFeed = feed;
		oldRPM = RPM;
		bInRapidMode = FALSE;
	    }

	    if (op->tl.use_css) {
		    MC_TRACE("WARNING: Ignored CSS feed %f max %d\n", fabs(n.u.m1.feed), op->tl.max_ss);
	    }

	    if (n.u.m2.ccomp > 0)	    // use this code to control surface normal curve
		MC_TRACE("WARNING: Cutter compensation ignored");

	    double dRad = sqrt(pow(n.u.m2.ep1[0] - n.u.m2.cpt[0], 2) +
			    pow(n.u.m2.ep1[1] - n.u.m2.cpt[1], 2) +
			    pow(n.u.m2.ep1[2] - n.u.m2.cpt[2], 2));
						
	    MC_TRACE("Comp = %d Plane = %d circle = %d position = %d", n.u.m2.ccomp, n.u.m2.pln, n.u.m2.circle, n.u.m2.position);
	    MC_TRACE("S%d, F%f", op->tl.rpm, n.u.m2.feed);
	    MC_TRACE("%s X%f Y%f Z%f CENTER(X%f Y%f Z%f) R%f", n.gcode == 2 ? "G2" : "G3", n.u.m2.ep1[0], n.u.m2.ep1[1], n.u.m2.ep1[2], n.u.m2.cpt[0], n.u.m2.cpt[1], n.u.m2.cpt[2], dRad);

	    if (n.u.m2.circle) { // full circle
		if (n.u.m2.pln == 0) {
		    double mid_x = ptStart[0] + 2 * (n.u.m2.cpt[0] - ptStart[0]);
		    double mid_y = ptStart[1] + 2 * (n.u.m2.cpt[1] - ptStart[1]);
		    double angle = atan2 (n.u.m2.ep1[1], n.u.m2.ep1[0]);
		    if (n.gcode == 2) // clockwsie
			angle = angle - PI / 3;
		    else   // anti-clockwise
			angle = angle + PI / 3;
		    double via_x = n.u.m2.cpt[0] + dRad * cos (angle);
		    double via_y = n.u.m2.cpt[1] + dRad * sin (angle);
		    goto_arc (path, via_x, via_y, n.u.m1.ep1[2]);
		    goto_arc (path, mid_x, mid_y, n.u.m1.ep1[2]);
		    goto_xyz (path, n.u.m1.ep1[0], n.u.m1.ep1[1], n.u.m1.ep1[2]);
		    MC_TRACE("Full circle on (XY) CENTER AXIS via_point at (X%f, Y%f) mid_point at (X%f, Y%f)", via_x, via_y, mid_x, mid_y);
		}
		else if (n.u.m2.pln == 1)
		    MC_TRACE("Full circle on (YZ) CENTER AXIS");
		else if (n.u.m2.pln == 3)
		    MC_TRACE("Full circle on (ZX) CENTER AXIS");
		else
		    MC_TRACE("Full circle on UNKNOWN CENTER AXIS!");
	    }
	    else if (n.u.m2.pln == 0) {
		// http://stackoverflow.com/questions/11674239/find-arcs-mid-point-given-start-end-and-center-of-circle-points
		double angle = atan2 (n.u.m2.ep1[1] *0.25 + ptStart[1]*0.75, n.u.m2.ep1[0]*0.25 + ptStart[0]*0.75);
		if (n.gcode == 2)   // switch to clockwise
		    angle = - angle;
		double via_x = n.u.m2.cpt[0] + dRad * cos (angle);
		double via_y = n.u.m2.cpt[1] + dRad * sin (angle);

		goto_arc (path, via_x, via_y, n.u.m1.ep1[2]);
		goto_xyz (path, n.u.m1.ep1[0], n.u.m1.ep1[1], n.u.m1.ep1[2]);
		MC_TRACE("ARC on (XY) CENTER AXIS via_point at (X%f, Y%f)", via_x, via_y);
	    }
	    else if (n.u.m2.pln == 1)
		MC_TRACE("ARC on (YZ) CENTER AXIS");
	    else if (n.u.m2.pln == 3)
		MC_TRACE("ARC on (ZX) CENTER AXIS");
	    else
		MC_TRACE("ARC on UNKNOWN CENTER AXIS!");

	    memcpy(ptStart, n.u.m2.ep1, 3 * sizeof(double));

	}

	else if (n.gcode == 81) {   // Drill (maybe first in a pattern
//	    MakeRoundHole(*opl, n.u.m0.ep1, g_Apt.GetCurrentWorkingstep());  // Code that made hole feature in DLL

	    p_3d ptTop, ptBot;					// hole top and bottom

	    memcpy(ptTop, n.u.m0.ep1, 3 * sizeof(double));
	    if (op->cmn.clearance_on)
		ptTop[Z] = ptTop[Z] - op->cmn.clearance_pln;
	    else if (op->cmn.retract_on) {  
		if (op->cmn.retract_inc)
		    ptTop[Z] = ptTop[Z] - op->cmn.retract_pln;
		else
		    ptTop[Z] = ptTop[Z];
	    }

	    memcpy(ptBot, ptTop, 3 * sizeof(double));
	    double depth = fabs(op->cmn.top_stock - op->cmn.depth);
	    MC_TRACE("Drilling Depth = %lf, break through = %lf", depth, op->u.prm_drl.brk_thru);
	    MC_TRACE("z = %lf Clearance = %lf, Retract = %lf feed_pln = %f", ptTop[Z], op->cmn.clearance_pln, op->cmn.retract_pln, op->cmn.feed_pln);
//	    depth = depth + op.u.op.u.prm_drl.brk_thru;	// KTH magnus says ignore

	    if (op->cmn.depth_inc) 
		ptBot[Z] = ptBot[Z] - depth;
	    else
		ptBot[Z] = op->cmn.depth;

	    // As per magnus feed/speed must change at the feed plane point
	    if (op->cmn.feed_inc) 
		ptTop[Z] = ptTop[Z] + op->cmn.feed_pln;			// this is the value called retract in UI
	    else
		ptTop[Z] = op->cmn.feed_pln;				// this is the value called retract in UI

	    // remember values for other points in pattern
	    hole_z_top = ptTop[Z];
	    hole_z_bot = ptBot[Z];

	    if (path) {
		MC_TRACE("Start point at (%lf, %lf, %lf)", ptTop[X], ptTop[Y], ptTop[Z]);
		goto_xyz (path, ptTop[0], ptTop[1], ptTop[2]);
	    }

	    path = new_toolpath (step_op, EN_TRAJECTORY_PATH, cmnt);
	    tech = find_or_make_technology (ws->getRootObject()->design(), op->tl.feed, feed_unit, RPM, speed_unit);
	    start_tool_tip_curve (path,  ptTop[0], ptTop[1], ptTop[2]);
	    set_toolpath_functions (path, funs);		    // funs set at top of loop
	    set_toolpath_technology (path, tech);
	    goto_xyz (path, ptBot[X], ptBot[Y], ptBot[Z]);
	    MC_TRACE("Bottom point at (%lf, %lf, %lf)", ptBot[X], ptBot[Y], ptBot[Z]);

	    path = new_toolpath_rapid (step_op, EN_TRAJECTORY_PATH, cmnt);
	    tech = find_or_make_technology (ws->getRootObject()->design(), op->tl.retract, feed_unit, RPM, speed_unit);
	    start_tool_tip_curve (path,  ptBot[0], ptBot[1], ptBot[2]);
	    set_toolpath_functions (path, funs);
	    set_toolpath_technology (path, tech);
	    goto_xyz (path, ptTop[X], ptTop[Y], ptTop[Z]);
	    MC_TRACE("Retract point at (%lf, %lf, %lf)", ptTop[X], ptTop[Y], ptTop[Z]);
	    MC_TRACE("Hole Depth = %f, spindle = %d feed = %f, plunge = %f, retract = %f", depth, RPM, op->tl.feed, op->tl.plunge, op->tl.retract);

	    oldFeed = op->tl.retract;
	    oldRPM = RPM;
	    bInRapidMode = FALSE;
	    memcpy(ptStart, ptTop, 3 * sizeof(double));
	}
	else if (n.gcode == 100) {// Second and subsequent in hole drilling pattern
	    p_3d ptTop, ptBot;					// hole top and bottom
	    ptTop[X] = n.u.m100.x; ptTop[Y] = n.u.m100.y; ptTop[Z] = hole_z_top;
	    ptBot[X] = n.u.m100.x; ptBot[Y] = n.u.m100.y; ptBot[Z] = hole_z_bot;

	    MC_TRACE("GOTO Drilling Next point at (%lf, %lf, %lf)", ptTop[X], ptTop[Y], ptTop[Z]);
	    // Continue at rapid pace from top point of last hole to start point for this one
	    goto_xyz (path, ptTop[0], ptTop[1], ptTop[2]);

	    path = new_toolpath (step_op, EN_TRAJECTORY_PATH, cmnt);
	    tech = find_or_make_technology (ws->getRootObject()->design(), op->tl.feed, feed_unit, RPM, speed_unit);
	    start_tool_tip_curve (path,  ptTop[0], ptTop[1], ptTop[2]);
	    set_toolpath_functions (path, funs);		    // funs set at top of loop
	    set_toolpath_technology (path, tech);
	    goto_xyz (path, ptBot[X], ptBot[Y], ptBot[Z]);
	    MC_TRACE("Bottom point at (%lf, %lf, %lf)", ptBot[X], ptBot[Y], ptBot[Z]);

	    path = new_toolpath_rapid (step_op, EN_TRAJECTORY_PATH, cmnt);
	    tech = find_or_make_technology (ws->getRootObject()->design(), op->tl.retract, feed_unit, RPM, speed_unit);
	    start_tool_tip_curve (path,  ptBot[0], ptBot[1], ptBot[2]);
	    set_toolpath_functions (path, funs);
	    set_toolpath_technology (path, tech);
	    goto_xyz (path, ptTop[X], ptTop[Y], ptTop[Z]);
	    MC_TRACE("Retract point at (%lf, %lf, %lf)", ptTop[X], ptTop[Y], ptTop[Z]);
	    MC_TRACE("Hole Depth = %f, spindle = %d feed = %f, plunge = %f, retract = %f", hole_z_top - hole_z_bot, RPM, op->tl.feed, op->tl.plunge, op->tl.retract);

	    oldFeed = op->tl.retract;
	    oldRPM = RPM;
	    bInRapidMode = FALSE;
	    memcpy(ptStart, ptTop, 3 * sizeof(double));
	}

    }

    return 0;
}

// Decode the Mastercam operation tree so can create workplan tree in STEP-NC
void GetOpParents(operation *opl, CStringArray *names)
{
    CStringArray reverse;

    //Get the immediate parent (a Toolpath Group) of this Operation
    op_group *op_parent = TpMainGrpMgr.GetMainGrpList().GroupByID(opl->cmn.grp_idn);
    CString parent(op_parent->name);
    reverse.Add(op_parent->name);			    // Add the name of the Toolpath Group 

    // Get the Machine Group that contains this Toolpath Group
    op_group *op_mach = TpMainGrpMgr.GetMainGrpList().RetrieveRootMachineGroup(op_parent->grp_idn);	
	
    int subLevel = 0;
    CString parentSub[6];
    long nextParent = op_parent->parent_grp_idn;
    while (nextParent != op_mach->grp_idn)
    {						
	op_parent = TpMainGrpMgr.GetMainGrpList().GroupByID(nextParent);
	reverse.Add(op_parent->name);			    // Add the name(s) of any Toolpath Groups that we're nested under
	parentSub[subLevel++] = op_parent->name;
	nextParent = op_parent->parent_grp_idn;
    }	
    reverse.Add(op_mach->name);			    // Add the Mchine Group name

    names->RemoveAll();
    for (INT_PTR i = reverse.GetCount() - 1; i >= 0; i--) {
	CString name = reverse.GetAt (i);
	names->Add (name);
    }
}